/************************************************************************/
/* 
 * Module:	cdctrl.c
 * Purpose:     to control Linux CD-ROMs
 * Language/Compiler/O/S:   GCC 2.7.2, Linux 2.0.X
 * Author:      Wade Hampton
 * Date:        12/22/97
 * Revision History:
 * [from RCS]
 *
 * $Log: cdctrl.c,v $
 * Revision 1.6  1998/07/15 13:49:58  wadeh
 * prior to Lin's changes
 *
 * Revision 1.5  1998/01/06 21:00:33  wadeh
 * Added DoCr, changes to fix cdeject crash.
 *
 * Revision 1.4  1997/12/30 22:59:31  wadeh
 * Added fflush to properly work with pipes...
 *
 * Revision 1.3  1997/12/24 15:39:10  wadeh
 * Minor bug fixes:
 * 1) Fixed cdctrl eject command causing a signal 11.
 * 2) Fixed Makefile debug option to not strip symbols, can now debug.
 * 3) Added command.h and more prototypes, more documentation.
 * 4) Fixed read_hw to always return buffer, even on error.
 * ,
 *
 * Revision 1.2  1997/12/23 22:52:51  wadeh
 * 2.1 beta 1 release
 *
 * Revision 1.1  1997/12/23 21:38:52  wadeh
 * Initial revision
 *
 *
 * Notes:
 * 1) 
 *----------------------------------------------------------------------*/
/* TYPES- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* INCLUDE FILES- - - - - - - - - - - - - - - - - - - - - - - - - - - - */
#include <stdio.h> 
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <linux/cdrom.h>

#include "config.h"
#include "cdctrl.h"
#include "commands.h"
#include "hardware.h"
#include "info.h"
#include "shuffle.h"
#include "util.h"

/* CONSTANTS- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
static char svnid[] = "$Id: cdctrl.c 665 2004-07-29 15:04:39Z max $";   

/* LOCAL DATA - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
char caLineIn[BUFSIZ];

/* PROTOTYPES - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void print_userhelp(int iVersion);
void print_help(void);
void print_info(cdtool_t *cdtool, int cdfd, char *pszCmd);
int  cmd_parse (cdtool_t *cdtool, int cdfd, char *pszBuf, int iRead);
/*
 * Copyright (C) 1997, 1998, Wade Hampton
 */
/************************************************************************/

/************************************************************************/
/* Procedure:  print_userhelp
 * Purpose:    to print command and user help on commands
 * 
 * Inputs:     none
 * Outputs:    to stdout, to stderr
 * Returns:    none
 * Notes:  
 *   1.
 */
/************************************************************************/
void print_userhelp(int iVersion)
{
    printf ("cdctrl:\nCommand line CDROM control\nOptions:\n");
    printf ("  cdctrl -c device  -- output using CRLF\n");
    printf ("  cdctrl device     -- normal output\n");
    printf ("  cdctrl [-h|-?|-V] -- help/version\n");

    if (iVersion == TRUE) {
	printf ("  Version %s\n", svnid);
	printf ("  Built on %s %s\n", __DATE__, __TIME__);
    } else {
	printf ("Command-line control options are:\n");
	print_help ();
    }
}

/************************************************************************/
/* Procedure:  print_help
 * Purpose:    to print help on commands
 * 
 * Inputs:     none
 * Outputs:    to stdout, to stderr
 * Returns:    none
 * Notes:  
 *   1.
 */
/************************************************************************/
void print_help ()
{
    INFOMSG("1    play (start play at track 1)");
    INFOMSG("s    stop");
    INFOMSG("p    pause");
    INFOMSG("r    resume");
    INFOMSG("e    eject");
    INFOMSG("c    close");
    INFOMSG("i    info");
    INFOMSG("d    dir");
    INFOMSG("l    shuffle (1 track only)");
    INFOMSG("-    previous track");
    INFOMSG("+    next track");
    INFOMSG("N    play track N");
    INFOMSG("q    quit");
    INFOMSG("?    help");
    INFOMSG("info format:  \"CMD cmd-name cd-status track abs-time rel-time\"");
    INFOMSG("  cmd-name  := {play, stop, ... quit} from above list");
    INFOMSG("  cd-status := {invalid, play, paused, completed, error, no status}");
    INFOMSG("  track     := {1..99} CD track");
    INFOMSG("  abs-time  := HH:MM:SS elapsed since CD start");
    INFOMSG("  rel-time  := HH:MM:SS elapsed since track start");
    INFOMSG("Build: %s %s [%s]", __DATE__, __TIME__, svnid);
}

/************************************************************************/
/* Procedure:  print_info
 * Purpose:    to print info on status
 * 
 * Inputs:     command, device of CDROM
 * Outputs:    to stdout, to stderr
 * Returns:    none
 * Notes:  
 *   1.
 */
/************************************************************************/
void print_info(cdtool_t *cdtool, int cdfd, char *pszCmd)
{
    printf ("CMD %s ", pszCmd);
    do_info (cdtool, cdfd, I_DISP_ALL); 
    do_crlf(stdout);
}

/************************************************************************/
/* Procedure:  cmd_parse
 * Purpose:    to parse user's commands
 * 
 * Inputs:     program name, file description for CDROM, buffer, 
 *             count, device of CDROM
 * Outputs:    to CDROM, to stdout, to stderr
 * Returns:    T/F, T when "done"
 * Notes:  
 *   1.
 */
/************************************************************************/
int cmd_parse (cdtool_t *cdtool, int cdfd, char *pszBuf, int iRead)
{
    int iDone = FALSE;
    int iTrk;
    char *pszCmd;

    pszCmd = "NULL";

    if ((pszBuf[0] >= '0') && (pszBuf[0] <= '9')) {
	iTrk = atoi (pszBuf);
	do_play(cdtool, cdfd, iTrk, 0, NULL);
	pszCmd = "play";
    } else {

	/* handle commands.... */
	switch (pszBuf[0]) {
	    
	case 's':
	    do_stop(cdtool, cdfd, 99);
	    pszCmd = "stop";
	    break;
	    
	case 'p':
	    do_pause(cdtool, cdfd);
	    pszCmd = "pause";
	    break;
	  
	case 'r':
	    do_play(cdtool, cdfd, 0, 0, NULL);
	    pszCmd = "resume";
	    break;
	  
	case 'e':
	    do_eject(cdtool, cdfd);
	    sleep (3);
	    pszCmd = "eject";
	    break;

	case 'c':
	    do_close(cdtool, cdfd);
	    sleep (3);
	    pszCmd = "close";
	    break;
	  
	case 'i':
        case 0x0d:
        case 0x0a:
	    pszCmd = "info";
	    break;
	    
	case 'd':
	    do_dir (cdtool, cdfd, P_RAW, 0); 
	    pszCmd = "dir";
	    break;

	case '-':
	    do_play(cdtool, cdfd, SKIPBACK, 0, NULL);
	    pszCmd = "previous";
	    break;

	case '+':
	    do_play(cdtool, cdfd, SKIPFORWARD, 0, NULL);
	    pszCmd = "next";
	    break;

	case 'q':
	    iDone = TRUE;
	    pszCmd = "QUIT";
	    break;

	case 'l':
	    do_shuffle (cdtool, cdfd, 99, 1, 1);
	    pszCmd = "shuffle";
	    break;

	case '?':
	    print_help();
	    pszCmd = "help";
	    break;


	default:
	    ERRORMSG("ERROR invalid command %s", pszBuf);
	    pszCmd = "default";
	    break;
	}
    }

    print_info(cdtool, cdfd, pszCmd);
    INFOMSG ("END");
    return (iDone);

}

/************************************************************************/
/* Procedure:  main
 * Purpose:    cdctrl's main 
 * 
 * Inputs:     command line args
 * Outputs:    to CDROM, to stdout, to stderr
 * Returns:    T/F, T when "done"
 * Notes:  
 *   1.
 */
/************************************************************************/
int main (int argc, char *argv[])
{
    cdtool_t cdtool;
    extern int optind, getopt();
    extern char *optarg;
    int cdfd;
    int opt=1;

    int iRead;
    int iDone = FALSE;

    cdtool.progname = "cdctrl";
    cdtool.cd_device = NULL;
    cdtool.cd_device_hint = NULL;

    cdtool_show_crlf = DOLF;
    cdtool_show_debug = FALSE;
    cdtool_show_verbose = FALSE;
    
    /* parse environment and command line args */
    if ((cdtool.cd_device_hint = getenv("CDTOOLDEV")) == NULL) 
      cdtool.cd_device_hint = strdup(CD_DEVICE);

    while (argv[opt] != NULL && argv[opt][0] == '-') {
      switch (argv[opt][1]) {    
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
  	  cdtool.cd_device_hint = strdup(CD_DEVICE "?");
  	  *(cdtool.cd_device_hint + sizeof(CD_DEVICE)-1) = argv[opt][1] - '0';
          argv++; argc--;
  	  break;
  
        case 'd':
          if (argv[opt+1] == NULL) {
              ERRORMSG("%s: -d needs an argument", cdtool.progname);
              exit(1);
          }
  	  cdtool.cd_device_hint = argv[opt+1];
          argv += 2; argc -= 2;
  	  break;
        
        case 'D':
  	  cdtool_show_debug = TRUE;
          argv++; argc--;
          break;
	
        case 'v':
  	  cdtool_show_verbose = TRUE;
          argv++; argc--;
          break;

        case 'c':
          cdtool_show_crlf = DOCRLF;
          argv++; argc--;
          break;
	  
        case 'h':
        case 'V':
        case '?':
          print_userhelp(FALSE);
          exit(0);
        
        default:
          ERRORMSG("%s: unknown option: %s", cdtool.progname, argv[opt]);
          exit(1);
      }
    }

    /* close stderr and redirect to stdout */
    close (2);
    dup2 (1, 2);
    
    /* try to find the CDROM */
    if ((cdtool.cd_device = device_find(cdtool.cd_device_hint)) == NULL) {
        ERRORMSG("%s: found no usable cdrom device.", cdtool.progname);
        print_userhelp(FALSE);
        exit(1);
    }
  
    /* try to open the CD device(s) */
    if ((cdfd = open(cdtool.cd_device, O_RDONLY | O_NONBLOCK)) == -1 ) {
	ERRORMSG("%s: can't open cdrom (%s)",cdtool.progname,cdtool.cd_device);
	exit(1);
    }
    
    INFOMSG("Opened %s as device %d", cdtool.cd_device, cdfd);

    /* process commands */
    while (iDone == FALSE) {
        if (fgets (caLineIn, sizeof(caLineIn)-1, stdin) != NULL) {
            iRead = strlen (caLineIn);
            if (iRead > 0) {
                iDone = cmd_parse (&cdtool, cdfd, caLineIn, iRead);
            }
        }
        fflush (stdout);
    }
    
    /* cleanup */
    close (cdfd);
    INFOMSG("EXITING");
    return (TRUE);
}

