/* "hardware.c" copyright 1994 thomas insel 
 *              copyrighy 1995 sven oliver moll */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>
#include <string.h>
#include <mntent.h>
#include <errno.h>

#ifdef sun
#  include <sundev/srreg.h>
#elif defined linux
#  include <linux/cdrom.h>
#else
#  error Please fix includes for your system in hardware.c
#endif

#include "config.h"
#include "database.h"
#include "hardware.h"
#include "util.h"

/* returns first track (1-based) on disk */
int
first_track(const cdhw_t *hw) {
  return hw->tochdr.cdth_trk0;
}

/* returns last track (1-based) on disk */
/* note that this is the last *real* track; a LEADOUT track follows (?) */
int
last_track(const cdhw_t *hw) {
  return hw->tochdr.cdth_trk1;
}

/* track is 1-based! */
int
is_audio_track(const cdhw_t *hw, int track) {
  return !(hw->tocentries[track - 1].cdte_ctrl & CDROM_DATA_TRACK);
}


/************************************************************************/
/* Procedure:  read_hw
 * Purpose:    to read hardware info
 * 
 * Inputs:     program name, CD file descriptor
 * Outputs:    to cdhw_t structure
 * Returns:    pointer to cdhw_t structure
 * Notes:  
 *   1.  The function free_hw_buf() must be called to free up the cdhw_t
 *       returned by this function.  This function uses malloc to get 
 *       a buffer (except when used with CDCTRL).
 */
/************************************************************************/
int
read_hw(cdtool_t *cdtool, cdhw_t **hwp, int cdfile)
{
  int ii;
  cdhw_t *hw = *hwp;

  /* WWH -- force all fields to 0 by default */
  memset (hw, 0, sizeof(cdhw_t));

  /* get status of CD-ROM */
  hw->iStatus = cd_status(cdtool, hw, cdfile);

  DEBUGMSG("read_hw:  buffer=0x%lx", (long)((void *)hw));
  DEBUGMSG("read_hw:  CD status=%s", cd_status_text(hw->iStatus));
  
  /* if mounted or no CD, return extra error codes */
  if (hw->iStatus == CD_STATUS_ERROR) {
      hw->subchnl.cdsc_audiostatus = CDROM_AUDIO_ERROR;
      return(-1);
  } else if ((hw->iStatus == CD_STATUS_NO_CD) ||
             (hw->iStatus == CD_STATUS_REMOVED)) {
      hw->subchnl.cdsc_audiostatus=NO_CDROM;
      return(-1);
  }
  else if (hw->iStatus == CD_STATUS_MOUNTED) {
      hw->subchnl.cdsc_audiostatus=CDROM_MOUNTED;
      return(-1);
  }

  /* at this point, status must be IN or INSERTED */
  /* read individual tracks */
  hw->iData = hw->iAudio = 0;
  for (ii=first_track(hw)-1; ii <= last_track(hw); ii++) {
      hw->tocentries[ii].cdte_format = CDROM_MSF;
      hw->tocentries[ii].cdte_track
          = (ii==last_track(hw)) ? CDROM_LEADOUT : ii+1;
      
      if (ioctl(cdfile, CDROMREADTOCENTRY, &(hw->tocentries[ii])) == -1) {
    	  ERRORMSG("%s: ioctl cdromreadtocentry track %d: %s", 
	    	  cdtool->progname, ii, strerror(errno));
    	  hw->subchnl.cdsc_audiostatus = CDROM_AUDIO_ERROR;
    	  return(-1);
      }
      
      /* determine if data or audio track, skip CDROM_LEADOUT */
      if (ii <= last_track(hw)) {
	  if (is_audio_track(hw, ii + 1))
	      hw->iAudio++;
	  else
	      hw->iData++;
      }
  }

  /* if no audio tracks, i.e., all-data disc, set status to DATA_DISC */
  if (hw->iAudio == 0) {
      hw->iStatus = CD_STATUS_DATA_DISC;
      hw->subchnl.cdsc_audiostatus = CDROM_DATA_DISC;
  } else {
      
      /* read subchannel info */
      hw->subchnl.cdsc_format = CDROM_MSF;
      if ( ioctl(cdfile, CDROMSUBCHNL, &(hw->subchnl)) == -1 ) {
	  ERRORMSG("%s: ioctl cdromsubchnl %s", cdtool->progname,
            strerror(errno));
          hw->subchnl.cdsc_audiostatus = CDROM_AUDIO_ERROR;
          return(-1);
      }
  }
  
  return(TRUE);
}

/************************************************************************/
/* Procedure:  checkmount 
 * Purpose:    to check if this CD is mounted
 * 
 * Inputs:     name of program, name of CD-ROM
 * Outputs:    debug
 * Returns:    T/F
 * Notes:  
 */
/************************************************************************/
int checkmount(cdtool_t *cdtool, char *pszName) {
  FILE  *fp;
  struct mntent *mnt;
  int    ii;
  char   caB[100];
  char   *pszTest;
  
  /* check pszName -- if NULL, return -1 */
  if (pszName == NULL)
      return (-1);

  /* find out if the device is a link, resolve link name */
  ii = readlink(pszName, caB, sizeof(caB)-1);
  if (ii < 0) {
    pszTest = pszName;
    DEBUGMSG("checkmount:  %s not a link", pszName);
  }else {
    pszTest = &caB[0];
    caB[ii] = '\0';
    DEBUGMSG("checkmount:  link -> %s", caB);
  }

  /* check if drive is mounted (from Mark Buckaway's cdplayer code) */
  if ((fp = setmntent (MOUNTED, "r")) == NULL) {
    ERRORMSG("%s: Couldn't open %s: %s", cdtool->progname, MOUNTED,
      strerror (errno));
    return (-1);
  }
  while ((mnt = getmntent (fp)) != NULL) {
    if (strstr (mnt->mnt_fsname, pszTest) != NULL) {
      endmntent (fp);
      return (TRUE);
    }
  }
  endmntent (fp);
  return (FALSE);
}

/************************************************************************/
/* Procedure:  cd_status_text
 * Purpose:    to get status name in text
 * 
 * Inputs:     status code
 * Outputs:    to buffer
 * Returns:    pointer to status text string
 * Notes:  
 *   1. Input values:
 *
 *	0	No CD in drive.
 *	1	CD in drive.
 *	2	CD has just been inserted (TOC has been read)
 *	3	CD has just been removed
 *	4	CD is mounted
 *
 */
/************************************************************************/
char *cd_status_text (int iStatus)
{
  char *pszResult;
  switch (iStatus) 
    {
    case CD_STATUS_ERROR:
      pszResult = "error";
      break;
    case CD_STATUS_NO_CD:
      pszResult = "no_disc";
      break;
    case CD_STATUS_CD_IN:
      pszResult = "disc_in";
      break;
    case CD_STATUS_INSERTED:
      pszResult = "inserted";
      break;
    case CD_STATUS_REMOVED:
      pszResult = "removed";
      break;
    case CD_STATUS_MOUNTED:
      pszResult = "mounted";
      break;
    case CD_STATUS_DATA_DISC:
      pszResult = "data_disc";
      break;
    default:
      pszResult = "unknown";
      break;
    }
  return (pszResult);
}

/************************************************************************/
/* Procedure:  cd_status
 * Purpose:    to get status of the CD-ROM
 * 
 * Inputs:     name of program (text)
 *             CD file descriptor (int), -1 if not open
 *             cdname (text) name or symlink to CD device
 *             pHW (cdhw_t) pointer to hardware descriptor block
 * Outputs:    debug info only
 * Returns:    see below
 * Notes:  
 *   1. Stolen from XPlaycd code...
 *   2. Return values:
 *
 *     -N       error code
 *	0	No CD in drive.
 *	1	CD in drive.
 *	2	CD has just been inserted (TOC has been read)
 *	3	CD has just been removed
 *	4	CD is mounted
 *
 */
/************************************************************************/
int cd_status(cdtool_t *cdtool, cdhw_t *pHw, int cd_fd) {
#if 0
  /* check if the CD-ROM is mounted, if so, return MOUNTED */
  /* disabled for now */
  iResult = checkmount(cdtool, cdname);
  if (iResult < 0)
    return (iResult);
  else if (iResult == TRUE)
    return CD_STATUS_MOUNTED;
#endif

  /* try to open the CD-ROM.  If we fail, report error */
  if (cd_fd < 0) {
    if ((cd_fd = open(cdtool->cd_device, O_RDONLY | O_NONBLOCK)) < 0) {
      if (errno == EACCES) {
	struct stat sb;
	struct group *g;
	gid_t groups[1024];
	int ngroups;
	
	ERRORMSG("%s: error opening %s: %s", cdtool->progname,
          cdtool->cd_device, strerror(errno));

	if (stat(cdtool->cd_device, &sb) < 0)
	  ERRORMSG("%s: cannot stat %s: %s", cdtool->progname,
            cdtool->cd_device, strerror(errno));
	else {
	  if (!(sb.st_mode & 0040)) {
	    ERRORMSG("%s is not group-readable.");
            ERRORMSG("Try, as root: chmod g+r %s",
              cdtool->cd_device, cdtool->cd_device);
          }

	  if (!(g = getgrgid(sb.st_gid))) {
	    ERRORMSG("%s is owned by unknown group %lu%s",
              cdtool->cd_device, (unsigned long)sb.st_gid);
            ERRORMSG("Try, as root: chgrp--dereference cdrom %s",
	      cdtool->cd_device);
          } else if (strcmp(g->gr_name, "cdrom")) {
	    ERRORMSG("%s is owned by group \"%s\"", 
              cdtool->cd_device, g->gr_name);
            ERRORMSG("Try, as root: chgrp --dereference cdrom %s",
              cdtool->cd_device);
          }
	}
	if (!(g = getgrnam("cdrom"))) {
	  ERRORMSG("This system appears to have no cdrom group!");
          ERRORMSG("Try, as root: addgroup cdrom");
        }
	else {
	  if ((ngroups = getgroups(sizeof groups/sizeof *groups, groups)) < 0)
	    ERRORMSG("%s: cannot get group list: %s",
              cdtool->progname, strerror(errno));
	  else {
	    int found = 0;
	    while (ngroups--) {
	      if (groups[ngroups] == g->gr_gid) {
		found = 1;
		break;
	      }
	    }
	    if (!found) {
	      struct passwd *p;
	      p = getpwuid(getuid());
	      ERRORMSG("You are not in the \"cdrom\" group.");
              ERRORMSG("Try, as root: adduser %s cdrom",
		p ? p->pw_name : "<your username>");
              ERRORMSG("You will need to log in again before this"
                "takes effect");
	    }
	  }
	}
      }
    }
    else if (errno == ENOENT) {
      ERRORMSG("%s: can't open %s: %s", cdtool->progname, cdtool->cd_device,
        strerror(errno));
      return (-1);
    }
    else if (errno != ENXIO) {
      return CD_STATUS_NO_CD;
    }
  }
  
  /* get subchannel info, if invalid, no disc in drive */  
  pHw->subchnl.cdsc_format = CDROM_MSF;
  if (ioctl(cd_fd, CDROMSUBCHNL, &pHw->subchnl)) 
    return CD_STATUS_NO_CD;

  /* read the header */
  if (ioctl(cd_fd, CDROMREADTOCHDR, &pHw->tochdr)) 
    return CD_STATUS_NO_CD;

  /* truncate too large number of tracks */
  if (pHw->tochdr.cdth_trk1 > MAX_TRACKS) {
      ERRORMSG("%s: reported track size is too large (%d > %d)",
        cdtool->progname, pHw->tochdr.cdth_trk1, MAX_TRACKS);
      pHw->tochdr.cdth_trk1 = MAX_TRACKS;
  }
  
  switch (pHw->subchnl.cdsc_audiostatus) {
  case CDROM_AUDIO_PLAY:
  case CDROM_AUDIO_PAUSED:
  case CDROM_AUDIO_COMPLETED:
    return CD_STATUS_CD_IN;
    break;
  }   

  return CD_STATUS_CD_IN;
}

/************************************************************************/
/* Procedure:  free_hw_buf
 * Purpose:    to free up hardware buffer returned by read_hw
 * 
 * Inputs:     pointer to buffer
 * Outputs:    debug info only
 * Returns:    n/a
 * Notes:  
 *   1.
 */
/************************************************************************/
void free_hw_buf(cdhw_t *pHw)
{
    DEBUGMSG("free_hw:  bufp=0x%lx", (long)((char *)pHw));

    if ((void *)pHw != NULL)
        free ((void *)pHw);
}
